==> stripe-3ds2-android-5.0.1-sources.jar: com/stripe/android/stripe3ds2/transaction/StripeTransaction.kt <==
package com.stripe.android.stripe3ds2.transaction

import android.app.Activity
import android.app.Dialog
import android.content.Intent
import com.nimbusds.jose.JOSEException
import com.stripe.android.stripe3ds2.exceptions.InvalidInputException
import com.stripe.android.stripe3ds2.init.ui.StripeUiCustomization
import com.stripe.android.stripe3ds2.observability.ErrorReporter
import com.stripe.android.stripe3ds2.security.MessageTransformer
import com.stripe.android.stripe3ds2.transactions.ChallengeRequestData
import com.stripe.android.stripe3ds2.transactions.ChallengeResponseData
import com.stripe.android.stripe3ds2.transactions.ErrorData
import com.stripe.android.stripe3ds2.views.ProgressViewFactory
import org.json.JSONException
import java.security.KeyPair
import java.security.PublicKey
import java.security.cert.CertificateException
import java.security.cert.X509Certificate
import java.text.ParseException

internal class StripeTransaction(
    private val areqParamsFactory: AuthenticationRequestParametersFactory,
    private val progressViewFactory: ProgressViewFactory,
    private val challengeStatusReceiverProvider: ChallengeStatusReceiverProvider,
    private val messageVersionRegistry: MessageVersionRegistry,
    private val sdkReferenceNumber: String,
    private val jwsValidator: JwsValidator,
    private val protocolErrorEventFactory: ProtocolErrorEventFactory,
    private val directoryServerId: String,
    private val directoryServerPublicKey: PublicKey,
    private val directoryServerKeyId: String?,
    override val sdkTransactionId: SdkTransactionId,
    private val sdkKeyPair: KeyPair,
    private val isLiveMode: Boolean,
    private val rootCerts: List<X509Certificate>,
    private val messageTransformer: MessageTransformer,
    private val uiCustomization: StripeUiCustomization?,
    private val brand: ProgressViewFactory.Brand,
    private val logger: Logger,
    private val errorReporter: ErrorReporter,
    private val challengeCompletionIntent: Intent? = null,
    private val challengeCompletionRequestCode: Int = 0
) : Transaction {
    override var initialChallengeUiType: String? = null

    private val acsDataParser = AcsDataParser(errorReporter)

    override suspend fun createAuthenticationRequestParameters(): AuthenticationRequestParameters {
        return areqParamsFactory.create(
            directoryServerId,
            directoryServerPublicKey,
            directoryServerKeyId,
            sdkTransactionId,
            sdkKeyPair.public
        )
    }

    override suspend fun doChallenge(
        activity: Activity,
        challengeParameters: ChallengeParameters,
        challengeStatusReceiver: ChallengeStatusReceiver,
        timeOut: Int
    ) {
        doChallenge(
            host = Stripe3ds2ActivityStarterHost(activity),
            challengeParameters = challengeParameters,
            challengeStatusReceiver = challengeStatusReceiver,
            timeOut = timeOut
        )
    }

    override suspend fun doChallenge(
        host: Stripe3ds2ActivityStarterHost,
        challengeParameters: ChallengeParameters,
        challengeStatusReceiver: ChallengeStatusReceiver,
        timeOut: Int
    ) {
        logger.info("Starting challenge flow.")

        val challengeCompletionIntentStarter = ChallengeCompletionIntentStarter.Default(
            host = host,
            requestCode = challengeCompletionRequestCode
        )

        runCatching {
            if (timeOut < MIN_TIMEOUT) {
                throw InvalidInputException("Timeout must be at least 5 minutes")
            }

            challengeStatusReceiverProvider.put(sdkTransactionId, challengeStatusReceiver)

            // will throw exception if acsSignedContent fails verification
            val (acsUrl, acsEphemPubKey) = getAcsData(
                requireNotNull(challengeParameters.acsSignedContent),
                isLiveMode,
                rootCerts
            )

            val creqData = createCreqData(challengeParameters)

            val errorRequestExecutor =
                StripeErrorRequestExecutor.Factory().create(acsUrl, errorReporter)

            // begin transaction timeout countdown
            val transactionTimer = TransactionTimer.Default(
                challengeStatusReceiver,
                timeOut,
                errorRequestExecutor,
                creqData,
                TransactionTimerProvider.Default
            ).also {
                it.start()
            }

            val creqExecutorConfig = ChallengeRequestExecutor.Config(
                messageTransformer,
                sdkReferenceNumber,
                sdkKeyPair.private.encoded,
                acsEphemPubKey.encoded,
                acsUrl,
                creqData
            )

            val result =
                StripeChallengeRequestExecutor.Factory()
                    .create(creqExecutorConfig, errorReporter)
                    .execute(creqData)

            onResult(
                result,
                host,
                creqExecutorConfig,
                challengeStatusReceiver,
                transactionTimer,
                errorRequestExecutor,
                challengeCompletionIntentStarter
            )
        }.onFailure {
            errorReporter.reportError(it)
            logger.error("Exception during challenge flow.", it)

            challengeStatusReceiver.runtimeError(
                RuntimeErrorEvent(it)
            ) {
                challengeCompletionIntent?.let { intent ->
                    challengeCompletionIntentStarter.start(
                        intent = intent,
                        outcome = ChallengeFlowOutcome.RuntimeError
                    )
                }
            }
        }
    }

    private fun onResult(
        result: ChallengeRequestResult,
        host: Stripe3ds2ActivityStarterHost,
        creqExecutorConfig: ChallengeRequestExecutor.Config,
        challengeStatusReceiver: ChallengeStatusReceiver,
        transactionTimer: TransactionTimer,
        errorRequestExecutor: ErrorRequestExecutor,
        challengeCompletionIntentStarter: ChallengeCompletionIntentStarter
    ) {
        when (result) {
            is ChallengeRequestResult.Success -> onResultSuccess(
                result.creqData,
                result.cresData,
                host,
                creqExecutorConfig
            )
            is ChallengeRequestResult.ProtocolError -> onProtocolErrorResult(
                result.data,
                challengeStatusReceiver,
                transactionTimer,
                errorRequestExecutor,
                challengeCompletionIntentStarter
            )
            is ChallengeRequestResult.RuntimeError -> onRuntimeErrorResult(
                result.throwable,
                challengeStatusReceiver,
                challengeCompletionIntentStarter
            )
            is ChallengeRequestResult.Timeout -> onTimeoutResult(
                result.data,
                challengeStatusReceiver,
                transactionTimer,
                errorRequestExecutor,
                challengeCompletionIntentStarter
            )
        }
    }

    private fun onResultSuccess(
        creqData: ChallengeRequestData,
        cresData: ChallengeResponseData,
        host: Stripe3ds2ActivityStarterHost,
        creqExecutorConfig: ChallengeRequestExecutor.Config
    ) {
        handleChallengeResponse(
            host,
            creqData,
            cresData,
            uiCustomization!!,
            creqExecutorConfig
        )
    }

    private fun onProtocolErrorResult(
        data: ErrorData,
        challengeStatusReceiver: ChallengeStatusReceiver,
        transactionTimer: TransactionTimer,
        errorRequestExecutor: ErrorRequestExecutor,
        challengeCompletionIntentStarter: ChallengeCompletionIntentStarter
    ) {
        transactionTimer.stop()
        sendErrorData(errorRequestExecutor, data)

        challengeStatusReceiver.protocolError(
            protocolErrorEventFactory.create(data)
        ) {
            startChallengeCompletionIntent(
                ChallengeFlowOutcome.ProtocolError,
                challengeCompletionIntentStarter
            )
        }
    }

    private fun onTimeoutResult(
        data: ErrorData,
        challengeStatusReceiver: ChallengeStatusReceiver,
        transactionTimer: TransactionTimer,
        errorRequestExecutor: ErrorRequestExecutor,
        challengeCompletionIntentStarter: ChallengeCompletionIntentStarter
    ) {
        transactionTimer.stop()
        sendErrorData(errorRequestExecutor, data)

        challengeStatusReceiver.runtimeError(RuntimeErrorEvent(data)) {
            startChallengeCompletionIntent(
                ChallengeFlowOutcome.RuntimeError,
                challengeCompletionIntentStarter
            )
        }
    }

    private fun onRuntimeErrorResult(
        throwable: Throwable,
        challengeStatusReceiver: ChallengeStatusReceiver,
        challengeCompletionIntentStarter: ChallengeCompletionIntentStarter
    ) {
        challengeStatusReceiver.runtimeError(RuntimeErrorEvent(throwable)) {
            startChallengeCompletionIntent(
                ChallengeFlowOutcome.RuntimeError,
                challengeCompletionIntentStarter
            )
        }
    }

    private fun startChallengeCompletionIntent(
        outcome: ChallengeFlowOutcome,
        challengeCompletionIntentStarter: ChallengeCompletionIntentStarter
    ) {
        challengeCompletionIntent?.let {
            challengeCompletionIntentStarter.start(
                intent = it,
                outcome = outcome
            )
        }
    }

    @Throws(InvalidInputException::class)
    override fun getProgressView(currentActivity: Activity): Dialog {
        return progressViewFactory.create(currentActivity, brand, uiCustomization!!)
    }

    override fun close() {}

    @Throws(
        ParseException::class, JOSEException::class, JSONException::class,
        CertificateException::class
    )
    private fun getAcsData(
        acsSignedContent: String,
        isLiveMode: Boolean,
        rootCerts: List<X509Certificate>
    ): AcsData {
        return acsDataParser.parse(
            jwsValidator.getPayload(acsSignedContent, isLiveMode, rootCerts)
        )
    }

    private fun handleChallengeResponse(
        host: Stripe3ds2ActivityStarterHost,
        creqData: ChallengeRequestData,
        cresData: ChallengeResponseData,
        uiCustomization: StripeUiCustomization,
        creqExecutorConfig: ChallengeRequestExecutor.Config
    ) {
        initialChallengeUiType = cresData.uiType?.code

        // start the initial challenge screen
        ChallengeStarter.create(
            host,
            creqData,
            cresData,
            uiCustomization,
            creqExecutorConfig,
            challengeCompletionIntent = challengeCompletionIntent,
            challengeCompletionRequestCode = challengeCompletionRequestCode
        ).start()
    }

    private fun sendErrorData(
        errorRequestExecutor: ErrorRequestExecutor,
        errorData: ErrorData
    ) {
        errorRequestExecutor.executeAsync(errorData)
    }

    private fun createCreqData(
        challengeParameters: ChallengeParameters
    ) = ChallengeRequestData(
        acsTransId = requireNotNull(challengeParameters.acsTransactionId),
        threeDsServerTransId = requireNotNull(challengeParameters.threeDsServerTransactionId),
        sdkTransId = sdkTransactionId,
        messageVersion = messageVersionRegistry.current
    )

    private companion object {
        private const val MIN_TIMEOUT = 5
    }
}

==> stripe-3ds2-android-5.1.0-sources.jar: com/stripe/android/stripe3ds2/transaction/StripeTransaction.kt <==
package com.stripe.android.stripe3ds2.transaction

import android.app.Activity
import android.app.Dialog
import com.nimbusds.jose.JOSEException
import com.stripe.android.stripe3ds2.exceptions.InvalidInputException
import com.stripe.android.stripe3ds2.init.ui.StripeUiCustomization
import com.stripe.android.stripe3ds2.observability.ErrorReporter
import com.stripe.android.stripe3ds2.security.MessageTransformer
import com.stripe.android.stripe3ds2.transactions.ChallengeRequestData
import com.stripe.android.stripe3ds2.transactions.ChallengeResponseData
import com.stripe.android.stripe3ds2.transactions.ErrorData
import com.stripe.android.stripe3ds2.views.ProgressViewFactory
import org.json.JSONException
import java.security.KeyPair
import java.security.PublicKey
import java.security.cert.CertificateException
import java.security.cert.X509Certificate
import java.text.ParseException

internal class StripeTransaction(
    private val areqParamsFactory: AuthenticationRequestParametersFactory,
    private val progressViewFactory: ProgressViewFactory,
    private val challengeStatusReceiverProvider: ChallengeStatusReceiverProvider,
    private val messageVersionRegistry: MessageVersionRegistry,
    private val sdkReferenceNumber: String,
    private val jwsValidator: JwsValidator,
    private val protocolErrorEventFactory: ProtocolErrorEventFactory,
    private val directoryServerId: String,
    private val directoryServerPublicKey: PublicKey,
    private val directoryServerKeyId: String?,
    override val sdkTransactionId: SdkTransactionId,
    private val sdkKeyPair: KeyPair,
    private val isLiveMode: Boolean,
    private val rootCerts: List<X509Certificate>,
    private val messageTransformer: MessageTransformer,
    private val uiCustomization: StripeUiCustomization?,
    private val brand: ProgressViewFactory.Brand,
    private val logger: Logger,
    private val errorReporter: ErrorReporter
) : Transaction {
    override var initialChallengeUiType: String? = null

    private val acsDataParser = AcsDataParser(errorReporter)

    override suspend fun createAuthenticationRequestParameters(): AuthenticationRequestParameters {
        return areqParamsFactory.create(
            directoryServerId,
            directoryServerPublicKey,
            directoryServerKeyId,
            sdkTransactionId,
            sdkKeyPair.public
        )
    }

    override suspend fun doChallenge(
        activity: Activity,
        challengeParameters: ChallengeParameters,
        challengeStatusReceiver: ChallengeStatusReceiver,
        timeOut: Int
    ) {
        doChallenge(
            host = Stripe3ds2ActivityStarterHost(activity),
            challengeParameters = challengeParameters,
            challengeStatusReceiver = challengeStatusReceiver,
            timeOut = timeOut
        )
    }

    override suspend fun doChallenge(
        host: Stripe3ds2ActivityStarterHost,
        challengeParameters: ChallengeParameters,
        challengeStatusReceiver: ChallengeStatusReceiver,
        timeOut: Int
    ) {
        logger.info("Starting challenge flow.")

        runCatching {
            if (timeOut < MIN_TIMEOUT) {
                throw InvalidInputException("Timeout must be at least 5 minutes")
            }

            challengeStatusReceiverProvider.put(sdkTransactionId, challengeStatusReceiver)

            // will throw exception if acsSignedContent fails verification
            val (acsUrl, acsEphemPubKey) = getAcsData(
                requireNotNull(challengeParameters.acsSignedContent),
                isLiveMode,
                rootCerts
            )

            val creqData = createCreqData(challengeParameters)

            val errorRequestExecutor =
                StripeErrorRequestExecutor.Factory().create(acsUrl, errorReporter)

            // begin transaction timeout countdown
            val transactionTimer = TransactionTimer.Default(
                challengeStatusReceiver,
                timeOut,
                errorRequestExecutor,
                creqData,
                TransactionTimerProvider.Default
            ).also {
                it.start()
            }

            val creqExecutorConfig = ChallengeRequestExecutor.Config(
                messageTransformer,
                sdkReferenceNumber,
                sdkKeyPair.private.encoded,
                acsEphemPubKey.encoded,
                acsUrl,
                creqData
            )

            val result =
                StripeChallengeRequestExecutor.Factory()
                    .create(creqExecutorConfig, errorReporter)
                    .execute(creqData)

            onResult(
                result,
                host,
                creqExecutorConfig,
                challengeStatusReceiver,
                transactionTimer,
                errorRequestExecutor
            )
        }.onFailure {
            errorReporter.reportError(it)
            logger.error("Exception during challenge flow.", it)

            challengeStatusReceiver.runtimeError(RuntimeErrorEvent(it))
        }
    }

    private fun onResult(
        result: ChallengeRequestResult,
        host: Stripe3ds2ActivityStarterHost,
        creqExecutorConfig: ChallengeRequestExecutor.Config,
        challengeStatusReceiver: ChallengeStatusReceiver,
        transactionTimer: TransactionTimer,
        errorRequestExecutor: ErrorRequestExecutor
    ) {
        when (result) {
            is ChallengeRequestResult.Success -> onResultSuccess(
                result.creqData,
                result.cresData,
                host,
                creqExecutorConfig
            )
            is ChallengeRequestResult.ProtocolError -> onProtocolErrorResult(
                result.data,
                challengeStatusReceiver,
                transactionTimer,
                errorRequestExecutor
            )
            is ChallengeRequestResult.RuntimeError -> onRuntimeErrorResult(
                result.throwable,
                challengeStatusReceiver
            )
            is ChallengeRequestResult.Timeout -> onTimeoutResult(
                result.data,
                challengeStatusReceiver,
                transactionTimer,
                errorRequestExecutor
            )
        }
    }

    private fun onResultSuccess(
        creqData: ChallengeRequestData,
        cresData: ChallengeResponseData,
        host: Stripe3ds2ActivityStarterHost,
        creqExecutorConfig: ChallengeRequestExecutor.Config
    ) {
        handleChallengeResponse(
            host,
            creqData,
            cresData,
            uiCustomization!!,
            creqExecutorConfig
        )
    }

    private fun onProtocolErrorResult(
        data: ErrorData,
        challengeStatusReceiver: ChallengeStatusReceiver,
        transactionTimer: TransactionTimer,
        errorRequestExecutor: ErrorRequestExecutor
    ) {
        transactionTimer.stop()
        sendErrorData(errorRequestExecutor, data)

        challengeStatusReceiver.protocolError(
            protocolErrorEventFactory.create(data)
        )
    }

    private fun onTimeoutResult(
        data: ErrorData,
        challengeStatusReceiver: ChallengeStatusReceiver,
        transactionTimer: TransactionTimer,
        errorRequestExecutor: ErrorRequestExecutor
    ) {
        transactionTimer.stop()
        sendErrorData(errorRequestExecutor, data)

        challengeStatusReceiver.runtimeError(RuntimeErrorEvent(data))
    }

    private fun onRuntimeErrorResult(
        throwable: Throwable,
        challengeStatusReceiver: ChallengeStatusReceiver
    ) {
        challengeStatusReceiver.runtimeError(
            RuntimeErrorEvent(throwable)
        )
    }

    @Throws(InvalidInputException::class)
    override fun getProgressView(currentActivity: Activity): Dialog {
        return progressViewFactory.create(currentActivity, brand, uiCustomization!!)
    }

    override fun close() {}

    @Throws(
        ParseException::class, JOSEException::class, JSONException::class,
        CertificateException::class
    )
    private fun getAcsData(
        acsSignedContent: String,
        isLiveMode: Boolean,
        rootCerts: List<X509Certificate>
    ): AcsData {
        return acsDataParser.parse(
            jwsValidator.getPayload(acsSignedContent, isLiveMode, rootCerts)
        )
    }

    private fun handleChallengeResponse(
        host: Stripe3ds2ActivityStarterHost,
        creqData: ChallengeRequestData,
        cresData: ChallengeResponseData,
        uiCustomization: StripeUiCustomization,
        creqExecutorConfig: ChallengeRequestExecutor.Config
    ) {
        initialChallengeUiType = cresData.uiType?.code

        // start the initial challenge screen
        ChallengeStarter(
            host,
            creqData,
            cresData,
            uiCustomization,
            creqExecutorConfig
        ).start()
    }

    private fun sendErrorData(
        errorRequestExecutor: ErrorRequestExecutor,
        errorData: ErrorData
    ) {
        errorRequestExecutor.executeAsync(errorData)
    }

    private fun createCreqData(
        challengeParameters: ChallengeParameters
    ) = ChallengeRequestData(
        acsTransId = requireNotNull(challengeParameters.acsTransactionId),
        threeDsServerTransId = requireNotNull(challengeParameters.threeDsServerTransactionId),
        sdkTransId = sdkTransactionId,
        messageVersion = messageVersionRegistry.current
    )

    private companion object {
        private const val MIN_TIMEOUT = 5
    }
}
